# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import re
from copy import deepcopy
# Logging
import logging
# Create logger
logger = logging.getLogger(__name__)
# All regular exceptions
REGEXP = re.compile(r'(.+?): ((.*))')
CPU_SYS_REG = re.compile(r'cpu[0-9]')
CPU_SYS_STATE_REG = re.compile(r'state[0-9]')
# Proc stat CPU usage
# https://www.linuxhowtos.org/System/procstat.htm
# cpu0 793125 280 352516 16192366 50291 0 2012 0 0 0
CPU_PROG_REG = re.compile(r'cpu(.+?) ((.*))')


def cpu_info():
    list_cpu = {}
    num_cpu = 0
    with open("/proc/cpuinfo", "r") as fp:
        for line in fp:
            # Search line
            match = REGEXP.search(line)
            if match:
                key = match.group(1).rstrip()
                value = match.group(2).rstrip()
                # Load value or if it is a new processor initialize a new field
                if key == "processor":
                    num_cpu = int(value)
                    list_cpu[num_cpu] = {}
                    continue
                # Load cpu info
                list_cpu[num_cpu][key] = value
    return list_cpu


# CPU lines
# - user: normal processes executing in user mode
# - nice: niced processes executing in user mode
# - system: processes executing in kernel mode
# - idle: twiddling thumbs
# - iowait: waiting for I/O to complete
# - irq: servicing interrupts
# - softirq: servicing softirqs
CPU_STAT_LABEL = ['user', 'nice', 'system', 'idle', 'iowait', 'irq', 'softirq', 'total']


def get_utilization(delta):
    if delta[-1] == 0:
        return {'user': 0, 'nice': 0, 'system': 0, 'idle': 0}
    # Return major outputs
    return {'user': 100.0 * (delta[0] / delta[-1]),
            'nice': 100.0 * (delta[1] / delta[-1]),
            'system': 100.0 * (delta[2] / delta[-1]),
            'idle': 100.0 * (delta[3] / delta[-1])
            }


def read_idle(path):
    # https://docs.kernel.org/admin-guide/pm/cpuidle.html
    states = [item for item in os.listdir(path) if os.path.isdir(os.path.join(path, item)) and CPU_SYS_STATE_REG.search(item)]
    idle = {}
    for state in sorted(states):
        with open("{path}/{state}/name".format(path=path, state=state), 'r') as f:
            name = f.read().strip()
        with open("{path}/{state}/disable".format(path=path, state=state), 'r') as f:
            disable = int(f.read())
        idle[name] = disable
    return idle


def read_freq_cpu(path, type_freq):
    # build dict freq
    freq = {}
    # Min frequency
    with open("{path}/cpufreq/{type_freq}_min_freq".format(path=path, type_freq=type_freq), 'r') as f:
        freq['min'] = int(f.read())
    # Max frequency
    with open("{path}/cpufreq/{type_freq}_max_freq".format(path=path, type_freq=type_freq), 'r') as f:
        freq['max'] = int(f.read())
    # Current frequency
    current_path = "{path}/cpufreq/{type_freq}_cur_freq".format(path=path, type_freq=type_freq)
    if os.path.isfile(current_path):
        with open(current_path, 'r') as f:
            data = f.read().strip()
            if data.isdigit():
                freq['cur'] = int(data)
    return freq


def read_system_cpu(path, cpu_status={}):
    # Online status
    cpu_status['online'] = True
    if os.path.isfile(path + "/online"):
        with open(path + "/online", 'r') as f:
            cpu_status['online'] = f.read().strip() == '1'
    # Read governor
    if os.path.isdir(path + "/cpufreq"):
        with open(path + "/cpufreq/scaling_governor", 'r') as f:
            cpu_status['governor'] = f.read().strip()
        # Store values
        cpu_status['freq'] = read_freq_cpu(path, 'scaling')
        cpu_status['info_freq'] = read_freq_cpu(path, 'cpuinfo')
    # Read idle CPU
    if os.path.isdir(path + "/cpuidle"):
        cpu_status['idle_state'] = read_idle(path + "/cpuidle")
    return cpu_status


class CPUService(object):

    def __init__(self):
        self._cpu_online = []
        # Load cpuinfo
        self._list_cpu = cpu_info()
        # List all CPU available
        self._proc_stat = "/proc/stat"
        path_system_cpu = "/sys/devices/system/cpu"
        if os.getenv('JTOP_TESTING', False):
            self._proc_stat = "/fake_sys/stat"
            path_system_cpu = "/fake_sys/devices/system/cpu"
            logger.warning("Running in JTOP_TESTING file={root_dir}".format(root_dir=self._proc_stat))
            logger.warning("Running in JTOP_TESTING folder={root_dir}".format(root_dir=path_system_cpu))
        if not os.path.isdir("/sys/devices/system/cpu"):
            raise Exception("Doesn't exist any CPU!")
        # Build a CPU list
        cpu_list = {int(item[3:]): {'path': "{path}/{item}".format(path=path_system_cpu, item=item),
                                    'last_cpu': [0.0] * len(CPU_STAT_LABEL),
                                    }
                    for item in os.listdir(path_system_cpu) if os.path.isdir(os.path.join(path_system_cpu, item)) and CPU_SYS_REG.search(item)}
        # Sort CPU list in a list by CPU name
        self._cpu = [cpu_list[i] for i in sorted(cpu_list)]
        # Status CPU service at start
        logger.info("Found {cpu} CPU".format(cpu=len(cpu_list)))
        # Build CPU total info
        self._cpu_total = {'last_cpu': [0.0] * len(CPU_STAT_LABEL)}
        # Check available cpufreq and cpuidle
        if not os.path.isdir(path_system_cpu + "/cpu0/cpufreq"):
            logger.warning("cpufreq folder not available on this device!")
        if not os.path.isdir(path_system_cpu + "/cpu0/cpuidle"):
            logger.warning("cpuidle folder not available on this device!")

    def reset_estimation(self):
        # reset estimation status cpu
        for cpu in self._cpu:
            cpu['last_cpu'] = [0.0] * len(CPU_STAT_LABEL)
        # Build CPU total info
        self._cpu_total = {'last_cpu': [0.0] * len(CPU_STAT_LABEL)}

    def get_cpu_info(self):
        # Load cpuinfo
        list_cpu = cpu_info()
        return [{'model': list_cpu.get(cpu, {}).get("model name", "")} for cpu in range(len(self._cpu))]

    def get_utilization(self, cpu_out):
        # CPU lines
        # - user: normal processes executing in user mode
        # - nice: niced processes executing in user mode
        # - system: processes executing in kernel mode
        # - idle: twiddling thumbs
        # - iowait: waiting for I/O to complete
        # - irq: servicing interrupts
        # - softirq: servicing softirqs
        total = {}
        with open(self._proc_stat, 'r') as f:
            for line in f:
                match = CPU_PROG_REG.search(line)
                if match:
                    # Get all fields
                    # 7 is equal to len(CPU_STAT_LABEL) - 1 (no total)
                    fields = [float(column) for column in match.group(2).strip().split()[:7]]
                    # Add total in last field
                    fields += [sum(fields)]
                    if match.group(1).isdigit():
                        num_cpu = int(match.group(1))
                        # Evaluate delta all values
                        # https://rosettacode.org/wiki/Linux_CPU_utilization
                        delta = [now - last for now, last in zip(fields, self._cpu[num_cpu]['last_cpu'])]
                        # Update last value
                        self._cpu[num_cpu]['last_cpu'] = deepcopy(fields)
                        # Store utilization
                        cpu_out[num_cpu].update(get_utilization(delta))
                    else:
                        # Evaluate delta all values
                        # https://rosettacode.org/wiki/Linux_CPU_utilization
                        delta = [now - last for now, last in zip(fields, self._cpu_total['last_cpu'])]
                        # Update last value
                        self._cpu_total['last_cpu'] = deepcopy(delta)
                        # Store utilization
                        total = get_utilization(delta)
                else:
                    # All CPU are in order on this file, if don't match we can skip it
                    break
        return total, cpu_out

    def get_status(self):
        # Status CPU
        cpu_list = [{} for i in range(len(self._cpu))]
        cpu_online = []
        # Add cpu status with frequency and idle config
        for cpu, data in enumerate(self._cpu):
            # store all data
            cpu_list[cpu] = read_system_cpu(data['path'], cpu_list[cpu])
            # Add model in CPU output
            cpu_list[cpu]['model'] = self._list_cpu.get(cpu, {}).get("model name", "")
            # Check status CPU
            cpu_online += [cpu_list[cpu]['online']]
        # Status number CPU changed if changed reset esimators
        # Fix weird TOTAL status when the number of CPU are changed (Numbers negative)
        for new, old in zip(cpu_online, self._cpu_online):
            if new != old:
                logger.info("Number of CPU online changed, reset estimators")
                self.reset_estimation()
                # reload cpuinfo
                self._list_cpu = cpu_info()
                break
        self._cpu_online = cpu_online
        # Usage CPU
        # https://stackoverflow.com/questions/8952462/cpu-usage-from-a-file-on-linux
        # https://www.linuxhowtos.org/System/procstat.htm
        # https://stackoverflow.com/questions/9229333/how-to-get-overall-cpu-usage-e-g-57-on-linux
        total, cpu_list = self.get_utilization(cpu_list)
        return {'total': total, 'cpu': cpu_list}
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
# Logging
import logging
# from .exceptions import JtopException
# Create logger
logger = logging.getLogger(__name__)


def read_engine(path):
    # Read status online
    engine = {}
    # Check if access to this file
    if os.access(path + "/clk_enable_count", os.R_OK):
        with open(path + "/clk_enable_count", 'r') as f:
            # Write online engine
            engine['online'] = int(f.read()) == 1
    # Check if access to this file
    if os.access(path + "/clk_rate", os.R_OK):
        with open(path + "/clk_rate", 'r') as f:
            # Write current engine
            engine['cur'] = int(f.read()) // 1000
    # Decode clock rate
    max_value = False
    if os.access(path + "/clk_max_rate", os.R_OK):
        with open(path + "/clk_max_rate", 'r') as f:
            # Write status engine
            value = int(f.read())
            # 18446744073709551615 = FFFF FFFF FFFF FFFF = 2 ^ 16
            if value != 18446744073709551615:
                engine['max'] = value // 1000
                max_value = True
    if os.access(path + "/clk_min_rate", os.R_OK) and max_value:
        with open(path + "/clk_min_rate", 'r') as f:
            # Write status engine
            engine['min'] = int(f.read()) // 1000
    return engine


class EngineService(object):

    ENGINES = ['ape', 'dla', 'pva', 'vic', 'nvjpg', 'nvenc', 'nvdec', 'se.', 'cvnas', 'msenc', 'ofa']

    def __init__(self):
        # Sort list before start
        EngineService.ENGINES.sort()
        self.engines_path = {}
        # List all engines available
        engine_path = "/sys/kernel/debug/clk"
        if os.getenv('JTOP_TESTING', False):
            engine_path = "/fake_sys/kernel/debug/clk"
            logger.warning("Running in JTOP_TESTING folder={root_dir}".format(root_dir=engine_path))
        list_all_engines = [x[0] for x in os.walk(engine_path)]
        # Search all available engines
        for name in EngineService.ENGINES:
            if name.endswith('.'):
                name = name[:-1]
                local_path = "{path}/{name}".format(path=engine_path, name=name)
                if os.path.isdir(local_path):
                    self.engines_path[name.upper()] = [local_path]
            else:
                # https://stackoverflow.com/questions/4843158/how-to-check-if-a-string-is-a-substring-of-items-in-a-list-of-strings
                local_path = "{path}/{name}".format(path=engine_path, name=name)
                # In this search are removed all engines that have a '.' on their name
                # like ape.buffer or nvdec.buf
                matching = [s for s in list_all_engines if local_path in s and '.' not in s]
                # Add in list all engines
                if matching:
                    # Check if name end with a number, if true collect by number
                    # dla0 dla1 ...
                    if os.path.basename(matching[0]).split('_')[0] == "{name}0".format(name=name):
                        logger.info("Special Engine group found: [{name}X]".format(name=name))
                        for num in range(10):
                            name_engine = "{name}{counter}".format(name=name, counter=num)
                            new_match = [match for match in matching if name_engine in match]
                            if new_match:
                                self.engines_path[name_engine.upper()] = sorted(new_match)
                            else:
                                break
                    else:
                        self.engines_path[name.upper()] = sorted(matching)
        # Print all engines found
        if self.engines_path:
            engines_string = ' '.join(name for name in self.engines_path)
            logger.info("Engines found: [{engines}]".format(engines=engines_string))
        else:
            logger.warn("Not engines found!")

    def get_status(self):
        status = {}
        # Read status from all engines
        for engine in self.engines_path:
            status[engine] = {}
            for local_path in self.engines_path[engine]:
                name_engine = os.path.basename(local_path).upper()
                logger.debug("Status [{engine}] in {path}".format(engine=name_engine, path=local_path))
                status[engine][name_engine] = read_engine(local_path)
        return status
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


from .common import cat, check_file
import os
import re
# Logging
import logging
# Create logger
logger = logging.getLogger(__name__)
TEMPERATURE_RE = re.compile(r'^temp(?P<num>\d+)_label$')
TEMPERATURE_OFFLINE = -256


def read_temperature(data):
    values = {}
    for name, path in data.items():
        try:
            value = float(cat(path)) / 1000.0
            values[name] = value
        except (OSError, ValueError):
            # If negative sensor offline
            values[name] = TEMPERATURE_OFFLINE
    return values


def get_virtual_thermal_temperature(thermal_path):
    temperature = {}
    # Find all thermal zone available
    items = os.listdir(thermal_path)
    subdirectories = ["{path}/{item}".format(path=thermal_path, item=item)
                      for item in items if os.path.isdir(os.path.join(thermal_path, item)) and 'thermal_' in item]
    # Check if is a real temperature
    idx = 0
    for thermal_path in subdirectories:
        path_name = "{path}/type".format(path=thermal_path)
        path_value = "{path}/temp".format(path=thermal_path)
        if os.path.isfile(path_name) and os.path.isfile(path_value):
            # Decode name
            raw_name = cat(path_name).strip()
            name = raw_name.split("-")[0] if '-' in raw_name else raw_name.split("_")[0]
            # Remove PMIC temperature (TX family)
            if 'PMIC' in name:
                logger.warn("Skipped PMIC")
                continue
            # Store new temperature
            name = name if name not in temperature else "{name}{idx}".format(name=name, idx=idx)
            idx = idx if name not in temperature else idx + 1
            # Check if is readable and accessible
            if check_file(os.path.join(thermal_path, "temp")):
                temperature[name] = {'temp': os.path.join(thermal_path, "temp")}
                # Message detected
                logger.info("Found thermal \"{name}\" in {path}".format(name=name, path=os.path.basename(thermal_path)))
    # Sort all temperatures
    return temperature


def get_hwmon_thermal_system(root_dir):
    sensor_name = {}
    if not os.path.isdir(root_dir):
        logger.error("Folder {root_dir} doesn't exist".format(root_dir=root_dir))
        return sensor_name
    # Fin all fans
    for dir in os.listdir(root_dir):
        path = os.path.join(root_dir, dir)
        if not os.path.isdir(path):
            continue
        # Find all pwm in folder
        for file in os.listdir(path):
            name_label_path = os.path.join(path, file)
            # Check if there is a label
            match = re.search(TEMPERATURE_RE, file)
            if not match:
                continue
            parsed_name = match.groupdict()
            # Build list current and average power read
            number_port = int(parsed_name['num'])
            # Read name
            raw_name = cat(name_label_path).strip()
            logger.info("Found temperature sensor: {name}".format(name=raw_name))
            # Build list of path
            path_crit_alarm = os.path.join(path, "temp{num}_crit_alarm".format(num=number_port))
            if check_file(path_crit_alarm):
                warnings = {'crit_alarm': path_crit_alarm}
                values = read_temperature(warnings)
                logger.info("Alarms {name} - {data}".format(name=raw_name, data=values))
            # Read Temperatures
            sensor = {}
            if check_file(os.path.join(path, "temp{num}_input".format(num=number_port))):
                sensor['temp'] = os.path.join(path, "temp{num}_input".format(num=number_port))  # Temperature in deg
            if check_file(os.path.join(path, "temp{num}_input".format(num=number_port))):
                sensor['max'] = os.path.join(path, "temp{num}_max".format(num=number_port))  # Temperature in deg
            if check_file(os.path.join(path, "temp{num}_input".format(num=number_port))):
                sensor['crit'] = os.path.join(path, "temp{num}_crit".format(num=number_port))  # Temperature in deg
            # If there is an file is added in list
            if sensor:
                sensor_name[raw_name] = sensor
    return sensor_name


class TemperatureService(object):

    def __init__(self):
        self._temperature = {}
        # Find all temperature available
        sys_folder = "/sys"
        if os.getenv('JTOP_TESTING', False):
            logger.warning("Running in JTOP_TESTING folder={root_dir}".format(root_dir=sys_folder))
        # Build folders
        hwmon_dir = os.path.join(sys_folder, "class", "hwmon")
        thermal_path = os.path.join(sys_folder, "devices", "virtual", "thermal")
        if os.path.isdir(thermal_path):
            # Sort all temperatures
            self._temperature = get_virtual_thermal_temperature(thermal_path)
        if os.path.isdir(hwmon_dir):
            hwmon_temperatures = get_hwmon_thermal_system(hwmon_dir)
            self._temperature.update(hwmon_temperatures)
        if not self._temperature:
            logger.warning("Temperature not folder found!")
        # Sort all sensors
        self._temperature = dict(sorted(self._temperature.items(), key=lambda item: item[0].lower()))

    def get_status(self):
        status = {}
        # Read temperature from board
        for name, sensor in self._temperature.items():
            values = read_temperature(sensor)
            # Status sensor
            values['online'] = values['temp'] != TEMPERATURE_OFFLINE
            # Add sensor in dictionary
            status[name] = values
        return status
# EOF
